// Code to maintain and search namespaces
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <idlc.h>
#include <parser.h>

void NameSpace::lookup_imports()
{
	for (list<StrListRef>::iterator i = import_strs.begin();
	     i != import_strs.end(); ++i)
	{
		StrList *strl = *i;
		Symbol *sym = lookup_sym(toplevel, strl, this);
		NameSpace *ns = dynamic_cast<NameSpace *>(sym);
		
		if (!ns) {
			const String *str = strl->back();
			yyerrorfl(str->file, str->line, "\"%s\" is not a namespace.",
			          sym->get_fq_name()->flatten()->c_str());
			throw UserError();
		}
		
		imports.push_back(ns);
	}

	for (const_iterator i = begin(); i != end(); ++i) {
		Symbol *sym = (*i).second;
		sym->lookup_imports();
	}
}


// Find the namespace in which "name" is declared.
// The rules for namespace precedence are in doc/idl/namespace-precedence.

NameSpace *NameSpace::search(const String *name, Symbol *exclude)
{	
	// Rule #1: Check current namespace first.
	Symbol *sym = lookup_noex(name, true);
	if (sym && sym != exclude)
		return this;

	// Rule #2: Check imported symbols
	Symbol *found = NULL;
	
	for (list<NameSpaceRef>::iterator i = imports.begin();
	     i != imports.end(); ++i)
	{
		NameSpace *ns = *i;
		Symbol *newfound = ns->lookup_noex(name, true);
		
		if (newfound) {
			if (newfound == exclude)
				continue;

			if (found && found != newfound) {
				yyerrorfl(name->file, name->line, 
				          "\"%s\" is ambiguous.  Two possibilities "
				          "(there may be more) are:",
				          name->c_str());

				yyerrorfl(found->name->file, found->name->line, "   \"%s\"",
				          found->get_fq_name()->flatten()->c_str());
				yyerrorfl(newfound->name->file, newfound->name->line, "   \"%s\"",
				          newfound->get_fq_name()->flatten()->c_str());

				throw UserError();
			}
			
			found = newfound;
		}
	}
	
	if (found)
		return found->get_ns();
	
	return NULL;
}

static NameSpace *search_for_namespace(const String *name, NameSpace *ctx,
                                       Symbol *exclude)
{
	while (ctx) {
		NameSpace *ret = ctx->search(name, exclude);

		if (ret)
			return ret;

		ctx = ctx->get_ns();
	}
	
	yyerrorfl(name->file, name->line,
	          "Unknown symbol \"%s\" in namespace search path.",
   	       name->c_str());

	throw UserError();
}

Symbol *lookup_sym(NameSpace *ns, StrList *name, NameSpace *ctx,
                   Symbol *exclude)
{
	Symbol *sym = NULL;
	NameSpace *top = ns;

	assert(current_pass != 1);
	assert(name->size() != 0);
	assert(ctx);
	
	StrList::iterator i = name->begin();
	
	if (name->front()->length() != 0) {
		ns = search_for_namespace(name->front(), ctx, exclude);
		
		if (!ns)
			ns = top;
	} else {
		++i;
	}

	assert(i != name->end());
	bool first = true;
	
	while (i != name->end()) {
		try {
			sym = ns->lookup(*i, first)->get_concrete_sym();
		}
		
		catch (SymbolNotFound) {
			yyerrorfl((*i)->file, (*i)->line, 
			          "Unknown symbol \"%s\" in \"%s\".",
			          (*i)->c_str(), ns->get_fq_name()->flatten()->c_str());

			throw UserError();
		}
		
		++i;
		first = false;

		ns = dynamic_cast<NameSpace *>(sym);

		if (!ns && i != name->end()) {
			yyerrorfl((*i)->file, (*i)->line, 
			          "\"%s\" is not a namespace.",
			          sym->get_fq_name()->flatten()->c_str());

			throw UserError();
		}
	}

	assert(sym);
	return sym;
}

// Returns the namespace (new or otherwise) on success, or NULL on failure.
NameSpace *add_nspace(StrList *name, bool push)
{
	StrList::iterator i = name->begin();
	NameSpace *ns = cur_nspace;
	Symbol *sym;
	
	if ((*i)->length() == 0) {
		yyerrorfl((*i)->file, (*i)->line, 
		          "Namespaces cannot be declared with an absolute path.");
		return NULL;
	}

	for (; i != name->end(); i++) {
		if ((*i)->token == '*') {
			yyerrorfl((*i)->file, (*i)->line, 
			          "'*' is only allowed with \"using\".");
			return NULL;
		}
	
		sym = ns->lookup_noex(*i, true);
		if (!sym) {
			UserNameSpace *new_ns;
			
			try {
				// Cannot throw DuplicateSymbol, but it can throw
				// InvalidArgument due to user error (such as trying
				// to implicitly declare a namespace inside of a
				// struct/class/whatever.
			
				new_ns = new UserNameSpace(*i);
				ns->add_user(new_ns);
			}
			
			catch (InvalidArgument) {
				yyerrorfl((*i)->file, (*i)->line,
				          "Cannot create namespace \"%s\" inside of a "
				          "non-namespace type or import namespace.",
				          (*i)->c_str());
				return NULL;
			}
			
			ns = new_ns;
		} else {
			// Don't let the user declare things in non-user namespaces. 
			// Besides the headache of verifying that it's something
			// that belongs, and of determining order for situations
			// where it matters, it's just icky.


			ns = dynamic_cast<UserNameSpace *>(sym);
			if (!ns) {
				yyerrorfl((*i)->file, (*i)->line,
				          "\"%s\" is not a namespace.",
				          sym->get_fq_name()->flatten()->c_str());
	
				return NULL;
			}
		}
	}
	
	if (push)
		nspace_stack.push_front(cur_nspace);

	cur_nspace = ns;
	return ns;
}

void pop_nspace()
{
	if (nspace_stack.empty())
		BUG();

	cur_nspace = nspace_stack.front();
	nspace_stack.pop_front();
}

void NameSpace::add_user(Symbol *sym)
{
	try {
		add(sym, false);
	}

	catch (DuplicateSymbol) {
		yyerrorfl(sym->name->file, sym->name->line, 
		          "\"%s\" already exists in this %s.",
		          sym->name->c_str(), description());
		throw UserError();
	}
}

void NameSpace::add_import(Symbol *sym, const char *filename)
{
	try {
		add(sym, true);
	}

	catch (DuplicateSymbol) {
		yyerrorf("\"%s\" already exists in %s.",
		         sym->name->c_str(), get_fq_name()->flatten()->c_str());
		BUG();
	}
	
	catch (InvalidArgument) {
		yyerrorf("\"%s\" caused an InvalidArgument upon add.", filename);
		throw UserError();
	}
}

Symbol::~Symbol()
{
	if (lang_priv)
		lang_priv->release();

	if (ns && !ns->dying) {
		try {
			ns->del(this);
		}
	
		catch (SymbolNotFound) {
			fprintf(stderr, "SymbolNotFound in Symbol::~Symbol(), cannot propagate\n");
		}
	}
}

StrList *Symbol::get_fq_name(const char *append, bool not_last) const
{
	StrList *ret;
	const String *s = NULL;

	if (name) {
		if (append && not_last &&
		    !dynamic_cast<const UserNameSpace *>(this))
		{
			String *mut_s = new String(name);
			mut_s->append(append);
			s = mut_s;
		} else {
			s = name;
		}
	} else if (!ns && !dynamic_cast<const UserNameSpace *>(this)) {
		s = new String("<temporary>");
	} else if (!ns) {
		s = new String("<toplevel>");
	} else {
		s = new String("<anonymous>");
	}
	
	if (!ns || !ns->name) {
		ret = new StrList();
	} else {
		ret = ns->get_fq_name(append, true);
	}

	ret->push_back(s);
	return ret;
}

Symbol *Symbol::find_toplevel_type()
{
	Symbol *cur = this, *prev;

	do {
		prev = cur;
		cur = cur->get_ns();
	} while (!dynamic_cast<UserNameSpace *>(cur));
	
	return prev;
}

// FIXME: Use double-dot for initial delimit, but only if
// generating IDLC-style names.  More generally, it should
// allow an arbitrary alternate global scope prefix.

String *StrList::flatten(const char *delimit)
{
	String *ret = new String();
	
	for (const_iterator i = begin(); i != end();) {
		const String *str = *i;
		ret->append(*str);
		
		if (++i != end())
			ret->append(delimit);
	}
	
	const String *str = back();
	ret->file = str->file;
	ret->line = str->line;
	
	return ret;
}
