// Code to output symbols to a legacy filesystem
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <cstring>
#include <cerrno>
#include <vector>
#include <memory>
#include <cctype>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <idlc.h>
#include <parser.h>

static string *map_sym(const char *root, Symbol *sym)
{
	typedef std::vector<StringRef> StringVec;
	StringVec path;
	
	while (sym && sym != output_ns && sym->name) {
		const String *name = NameSpace::mangle(sym->name);
		path.push_back(name);
		sym = sym->get_ns();
	}
	
	return stringvec_to_path(path, root);
}

string *stringvec_to_path(StringVec &stringvec, const char *prepend)
{
	string *s = new string();
	
	if (prepend)
		*s += string(prepend);

	for (StringVec::reverse_iterator i = stringvec.rbegin();
	     i != stringvec.rend(); ++i)
	{
		*s += '/';
		*s += *i;
	}
	
	return s;
}

void NameSpace::output(const char *root)
{
	std::auto_ptr<string> path(map_sym(root, this));

	if (mkdir(path->c_str(), 0777) && errno != EEXIST) {
		fprintf(stderr, "Cannot create directory \"%s\": %s.\n",
		        path->c_str(), strerror(errno));

		throw UserError();
	}

	for (tbl_type::iterator i = tbl.begin(); i != tbl.end(); ++i) {
		Symbol *sym = (*i).second;

		if (!sym->priv)
			sym->output(root);
	}
}

void Def::output_self(const char *root, Symbol *sym, bool dir)
{
	std::auto_ptr<string> path(map_sym(root, sym));

	if (dir)
		*path += "/.self";
	
	int fd = open(path->c_str(), O_CREAT | O_EXCL | O_WRONLY, 0666);
	if (fd < 0) {
		fprintf(stderr, "Cannot create file \"%s\": %s.\n",
		        path->c_str(), strerror(errno));

		throw UserError();
	}
	
	// open() is used so that O_EXCL can be specified, but 
	// fwrite is used for buffering.
	File f = fdopen(fd, "wb");
	if (!f) {
		fprintf(stderr, "Cannot fdopen \"%s\": %s.\n",
		        path->c_str(), strerror(errno));

		BUG();
	}

	if (fwrite(&hdr, sizeof(hdr), 1, f) != 1)
		goto err;
	if (fwrite(self, self_len, 1, f) != 1)
		goto err;
	if (!output_extra(f))
		goto err;

	return;
	
err:
	fprintf(stderr, "Cannot write to file \"%s\": %s.\n",
	        path->c_str(), strerror(ferror(f)));

	throw UserError();
}

template<typename T, bool namespace_qual>
bool output_list(list<T> &entries, int count, FILE *f)
{
	int index = 0;

	for (typename list<T>::iterator i = entries.begin();
	     i != entries.end(); ++i)
	{
		StringRef s;
		
		if (index >= count)
			BUG();
		
		if (namespace_qual)
			s = (*i)->get_fq_name()->flatten();
		else
			s = (*i)->name;
			
		int32_t len = s->length();

		if (fwrite(&len, sizeof(len), 1, f) != 1)
			return false;
		
		if (fwrite(s->c_str(), len + 1, 1, f) != 1)
			return false;
		
		// Make sure the next string's byte count is aligned on a
		// 4-byte boundary.  Len now includes the null.
		
		len++;

		int pad = ((len + 3) & ~3) - len;
		int32_t zero = 0;
		
		if (pad && fwrite(&zero, pad, 1, f) != 1)
			return false;
		
		index++;
	}
	
	if (index != count)
		BUG();
	
	return true;
}

void Enum::output(const char *root)
{
	NameSpace::output(root);
	output_self(root, this, true);
}

bool Enum::output_extra(FILE *f)
{
	return output_list<DatumRef, false>(entries, def.num_entries, f);
}

void Interface::output(const char *root)
{
	NameSpace::output(root);
	output_self(root, this, true);
}

bool Interface::output_extra(FILE *f)
{
	return output_list<MethodRef, false>(methods, def.num_methods, f) &&
	       output_list<InterfaceRef, true>(supers, def.num_supers, f);	
}

void Method::output(const char *root)
{
	NameSpace::output(root);
	output_self(root, this, true);
}

bool Method::output_extra(FILE *f)
{
	return output_list<ParamRef, false>(entries, def.num_entries, f);
}

void Param::output(const char *root)
{
	output_self(root, this, false);
}

bool Param::output_extra(FILE *f)
{
	if (!basic)
		return fwrite(type_fq_name->c_str(), def.type.length + 1, 1, f) == 1;
	else
		return true;
}

void Struct::output(const char *root)
{
	NameSpace::output(root);
	output_self(root, this, true);
}

bool Struct::output_extra(FILE *f)
{
	if (!output_list<DatumRef, false>(entries, def.num_entries, f))
		return false;

	if (super) {
		list<Struct *> superlist;
		superlist.push_back(super);
		return output_list<Struct *, true>(superlist, 1, f);
	}
	
	return true;
}

void BitField::output(const char *root)
{
	NameSpace::output(root);
	output_self(root, this, true);
}

bool BitField::output_extra(FILE *f)
{
	return output_list<DatumRef, false>(entries, def.num_entries, f);
}

void Datum::output(const char *root)
{
	if (def.flags.field.Const && !const_init) {
		resolve_constant_chain();
	
		if (!const_init)
			throw UserError();
	}

	output_self(root, this, false);
}

bool Datum::output_extra(FILE *f)
{
	if (!basic)
		return fwrite(type_fq_name->c_str(), def.type.length + 1, 1, f) == 1;
	else
		return true;
}

void BasicType::output(const char *root)
{
	if (!complete)
		BUG();

	output_self(root, this, false);
}

void Alias::output(const char *root)
{
	output_self(root, this, false);
}

bool Alias::output_extra(FILE *f)
{
	return fwrite(sym_fq_name->c_str(), def.length + 1, 1, f) == 1;
}

void UserNameSpace::output(const char *root)
{
	NameSpace::output(root);
	output_self(root, this, true);
}

bool UserNameSpace::output_extra(FILE *f)
{
	return fwrite(mountpoint_name->c_str(), def.length + 1, 1, f) == 1;
}
