// Semantic actions for types and namespaces
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <cfloat>
#include <sstream>

#include <idlc.h>
#include <parser.h>
#include <lang.h>

bool Datum::verify_const()
{
	assert(def.flags.field.Const);
	
	// TOK_INVALID should have caused an abort in an earlier pass, and
	// TOK_DCON should have been replaced by the final constant by now.
	
	assert(con_type == TOK_ICON || con_type == TOK_UCON || 
	       con_type == TOK_FCON || con_type == TOK_BOOL);

	if (cbt->flags.field.Float) {
		if (con_type == TOK_UCON)
			def.fcon = static_cast<double>(def.ucon);
		else if (con_type == TOK_ICON)
			def.fcon = static_cast<double>(def.icon);
		else if (con_type == TOK_BOOL) {
			yyerrorfl(name->file, name->line, 
			          "%s is of floating-point type, but is initialized with "
			          "a boolean constant.", name->c_str());

			return false;
		}

		con_type = TOK_FCON;
		
		if (cbt->bits == 32) {
			if (def.fcon > FLT_MAX || def.fcon < FLT_MIN) {
				yyerrorfl(name->file, name->line, 
				          "%s is out of range for a single-precision FP datum.",
				          name->c_str());
				
				return false;
			}
		}
		
		const_init = true;
		return true;
	}

	if (con_type == TOK_FCON) {
		yyerrorfl(name->file, name->line, 
		          "%s is of %s type, but is initialized with "
		          "a floating-point constant.", 
		          cbt->flags.field.Bool ? "boolean" : "integral",
		          name->c_str());

		return false;
	}
	
	if (cbt->flags.field.Bool) {
		if (con_type != TOK_BOOL) {
			yyerrorfl(name->file, name->line, 
			          "%s is of boolean type, but is initialized with "
			          "a non-boolean constant.", name->c_str());

			return false;
		}
		
		assert(def.ucon == 0 || def.ucon == 1);
		
		const_init = true;
		return true;
	}

	if (con_type == TOK_BOOL) {
		yyerrorfl(name->file, name->line, 
		          "%s is of integral type, but is initialized with "
		          "a boolean constant.", name->c_str());

		return false;
	}

	if (cbt->flags.field.Unsigned) {
		if (con_type == TOK_ICON && def.icon < 0) {
			yyerrorfl(name->file, name->line, 
			          "Initializer for \"%s\" is out of range.",
			          name->c_str());
			return false;
		}

		if (cbt->bits < 64) {
			uint64_t max = (1ULL << cbt->bits) - 1;
			
			if (def.ucon > max) {
				yyerrorfl(name->file, name->line, 
				          "Initializer for \"%s\" is out of range.",
				          name->c_str());
				return false;
			}
		}
	} else {
		if (con_type == TOK_UCON && def.icon < 0) {
			yyerrorfl(name->file, name->line, 
			          "Initializer for \"%s\" is out of range.",
			          name->c_str());
			return false;
		}
		
		if (cbt->bits < 64) {
			// Yes, the constant should be unsigned before conversion,
			// as the temporary value could overflow a signed datum
			// before the one is subtracted if bits == 63.  It won't
			// matter on most machines, but you never know...
			
			int64_t max = (1ULL << (cbt->bits - 1)) - 1;
			
			if (def.icon > max || def.icon < -max + 1) {
				yyerrorfl(name->file, name->line, 
				          "Initializer for \"%s\" is out of range.",
				          name->c_str());
				return false;
			}
		}
	}

	const_init = true;
	return true;
}

void Datum::init_const_early(Con *con)
{
	if (con) {
		if (con->type == TOK_DCON) {
			assert(current_pass == 1);
			const_val_name = con->con.dcon;
		}

		def.flags.field.Const = 1;
		con_type = con->type;
		def.ucon = con->con.ucon;
	}
}

void Datum::init(StrList *type, Array *ARRAY, Con *con)
{
	assert(!complete);
	
	type_name = type;
	cbt = NULL;
	
	basic = false;
	complete = true;

	set_array(ARRAY);
	init_const_early(con);
}

void Datum::init(CompiledBasicType &CBT, Array *ARRAY, Con *con)
{
	assert(!complete);
	
	def.basictype = CBT;
	cbt = &def.basictype;

	basic = true;
	complete = true;
	
	set_array(ARRAY);
	init_const_early(con);
}

void Datum::process_type()
{
	assert(current_pass >= 4);

	if (cbt)
		return;
	
	assert(!type_fq_name);
	assert(type_sym);
	
	Symbol *sym = type_sym->get_concrete_sym(false);
	type = dynamic_cast<Type *>(sym);
	
	if (!type) {
		const String *str = type_name->back();
		yyerrorfl(str->file, str->line, "\"%s\" is not a type.",
		          sym->get_fq_name()->flatten()->c_str());

		throw UserError();
	}
	
	BasicType *bt = dynamic_cast<BasicType *>(type->get_concrete_sym());
	
	if (bt) {
		if (!bt->def.flags.field.TypeDef)
			use_anon_type(bt->def);
		else
			cbt = &bt->def;
	} else if (const_val) {
		const String *str = type_name->back();
		yyerrorfl(str->file, str->line, 
		          "\"%s\" is not a basic type (and thus \"%s\" cannot "
		          "be const).",
		          type->get_fq_name()->flatten()->c_str(),
		          name->c_str());
		throw UserError();
	}
	
	if (!basic) {
		type_fq_name = type->get_fq_name()->flatten();
		def.type.length = type_fq_name->length();
	}
}

Datum *Datum::resolve_constant_chain()
{
	if (chain_traversed == traversal) {
		yyerrorfl(name->file, name->line, 
		          "Initialization of \"%s\" forms an infinite loop:",
		          get_fq_name()->flatten()->c_str());
		return this;
	}

	chain_traversed = traversal;
	
	assert(!const_val);
	assert(!const_init);

	process_type();

	if (const_val_sym) {
		const_val = dynamic_cast<Datum *>(const_val_sym->get_concrete_sym());
		
		if (!const_val) {
			const String *str = const_val_name->back();
			yyerrorfl(str->file, str->line, "\"%s\" is not a datum.",
			          const_val_sym->get_fq_name()->flatten()->c_str());
			throw UserError();
		}
		
		if (!const_val->def.flags.field.Const) {
			const String *str = const_val_name->back();
			yyerrorfl(str->file, str->line, "\"%s\" is not a constant.",
			          const_val->get_fq_name()->flatten()->c_str());
			throw UserError();
		}
	
		if (!const_val->const_init) {
			Datum *d = const_val->resolve_constant_chain();
			if (d) {
				yyerrorfl(name->file, name->line, "...initializes \"%s\"",
				          get_fq_name()->flatten()->c_str());
		
				if (d == this)
					throw UserError();
				
				return d;
			}
		}
		
		assert(const_val->const_init);
	
		con_type = const_val->con_type;
		def.ucon = const_val->def.ucon;
	}

	if (!verify_const())
		throw UserError();

	assert(const_init);
	return NULL;
}

// FIXME: Check for inline struct loops.
void Datum::final_analysis()
{
	Struct *str = dynamic_cast<Struct *>(*type);
	if (str && str->is_inline())
		set_inline();

	if (array) {
		array->final_analysis();
		def.basictype.array = array->ca;
	} else {
		def.basictype.array.bounds[0] = 0;
		def.basictype.array.bounds[1] = 0;
	}
	
	BitField *bfparent = dynamic_cast<BitField *>(get_ns());
	if (bfparent) {
		assert(!array);
		int defsize;
		
		if (basic)
			defsize = 1;
		else
			defsize = type->get_default_bf_size();
		
		if (defsize == -1) {
			yyerrorfl(name->file, name->line,
			          "\"%s\" may not be the named type of bitfield entry "
			          "\"%s\", as it is not a bitfield or enumeration.",
			          type->get_fq_name()->flatten()->c_str(),
			          get_fq_name()->flatten()->c_str());
		} else if (def.icon == -1) {
			def.ucon = defsize;
		} else if (def.ucon < (unsigned int)defsize) {
			yyerrorfl(name->file, name->line,
			          "Explicit size %llu on \"%s\" is too small for "
			          "type \"%s\", which has a minimum size of %u",
			          (unsigned long long)def.ucon,
			          get_fq_name()->flatten()->c_str(),
			          type->get_fq_name()->flatten()->c_str(), defsize);
		}
	}
}

void Struct::add(Symbol *sym, bool from_import)
{
	Datum *datum = dynamic_cast<Datum *>(sym);
	
	if (!datum && !dynamic_cast<Alias *>(sym) && 
	    !dynamic_cast<Type *>(sym) &&
	    !dynamic_cast<Method *>(sym))
		throw InvalidArgument();
	
	NameSpace::add(sym, from_import);

	if (!from_import && datum && !datum->def.flags.field.Const)
		add_elem(datum);
}

void Struct::add_elem(Datum *d)
{
	entries.push_back(d);
	def.num_entries++;
}

bool Struct::is_plain_data()
{
	if (!plain_data) {
		plain_data = 1;
		for (Struct::entries_iterator i = entries_begin(); i != entries_end(); ++i)
		{
			Datum *d = *i;
			if (d->is_array() && !d->def.flags.field.Inline)
				goto complex;

			if (!d->type)
				continue;

			Type *type = dynamic_cast<Type *>(d->type->get_concrete_sym());
			assert(type);

			if (dynamic_cast<Interface *>(type))
				goto complex;

			Struct *str = dynamic_cast<Struct *>(type);
			if (str) {
				if (!d->def.flags.field.Inline)
					goto complex;

				// FIXME: check for inline struct loops
				if (str->is_plain_data())
					goto complex;
			}
		}
	}
	
	assert(plain_data == 1 || plain_data == 2);
	return plain_data == 1;

complex:
	plain_data = 2;
	return false;
}

Param *Param::declare(const String *name, NameSpace *parent,
	                   StrList *TYPE, CompiledParam::Flags flags,
	                   Array *array)
{
	assert(parent);
	assert(dynamic_cast<Method *>(parent));
	
	Param *p = new Param(name);
	p->init(TYPE, flags, array);
	
	parent->add_user(p);
	return p;
}

Method *Method::declare(const String *name, NameSpace *parent)
{
	assert(parent);
	assert(dynamic_cast<Interface *>(parent));

	Method *m = new Method(name);

	parent->add_user(m);
	return m;
}

void Interface::add(Symbol *sym, bool from_import)
{
	Datum *datum = dynamic_cast<Datum *>(sym);
	Method *method = dynamic_cast<Method *>(sym);

	if (!datum && !method && 
	    !dynamic_cast<Alias *>(sym) && 
	    !dynamic_cast<Type *>(sym))
		throw InvalidArgument();

	if (datum && !datum->def.flags.field.Const)
		throw InvalidArgument();
	
	NameSpace::add(sym, from_import);
	
	if (!from_import && method)
		add_elem(method);
}

// FIXME: check for duplicate and looping inheritance
void Interface::add_elem(Method *method)
{
	methods.push_back(method);
	def.num_methods++;
}

void Enum::add(Symbol *sym, bool from_import)
{
	Datum *datum = dynamic_cast<Datum *>(sym);

	// It also must be const unsigned, but the const won't
	// have been initialized yet.

	if (!datum)
		throw InvalidArgument();
	
	NameSpace::add(sym, from_import);
	
	if (!from_import)
		add_elem(datum);
}

void Enum::add_elem(Datum *datum)
{
	entries.push_back(datum);
	def.num_entries++;
}

void BitField::add(Symbol *sym, bool from_import)
{
	Datum *datum = dynamic_cast<Datum *>(sym);

	if (!datum && !dynamic_cast<Enum *>(sym) &&
	    !dynamic_cast<BitField *>(sym))
		throw InvalidArgument();

	NameSpace::add(sym, from_import);

	if (!from_import && datum)
		add_elem(datum);
}

void BitField::add_elem(Datum *datum)
{
	entries.push_back(datum);
	def.num_entries++;
}

void Method::add(Symbol *sym, bool from_import)
{
	Param *param = dynamic_cast<Param *>(sym);

	if (!param)
		throw InvalidArgument();

	NameSpace::add(sym, from_import);
	if (!from_import)
		add_elem(param);
}

void Method::add_elem(Param *param)
{
	entries.push_back(param);
	def.num_entries++;
}

// Use this rather than lookup_sym if you want to check for built-in types.
// If only constructed types are valid, use lookup_sym.  If basic_types_only
// is set, then NULL will be returned if sl does not describe a built-in
// basic type (such as int, flong, octet, etc).  This option is used so that
// the first-pass code in idlparse.y can decide whether to create a BasicType
// or an Alias/TypeDef.

Type *lookup_type(StrList *sl, NameSpace *ctx, bool basic_types_only)
{
	Type *type;
	int token = sl->front()->token;

	// These tokens aren't valid types, but the names can be
	// used as identifiers.
	switch (token) {
		case TOK_ASYNC:
		case TOK_INOUT:
		case TOK_OUT:
			token = TOK_IDENT;
	}
	
	// Treat it as a user-defined type if it's either not a reserved word,
	// or if there are multiple components.

	if (token == TOK_IDENT || sl->front() != sl->back()) {
		if (basic_types_only)
			return false;
	
		Symbol *sym = lookup_sym(toplevel, sl, ctx);
		type = dynamic_cast<Type *>(sym->get_concrete_sym(false));
		
		if (!type) {
			const String *str = sl->back();
			yyerrorfl(str->file, str->line, "\"%s\" is not a type.",
			          sym->get_fq_name()->flatten()->c_str());

			throw UserError();
		}
	} else {
		CompiledBasicType cbt;
		memset(&cbt, 0, sizeof(cbt));
		
		switch (token) {
			case TOK_BOOL:
				cbt.bits = 0;
				cbt.flags.field.Bool = 1;
				break;
			
			case TOK_OCTET:
			case TOK_CHAR:
				cbt.bits = 8;
				cbt.flags.field.Unsigned = 1;
				break;
			
			case TOK_SHORT:
				cbt.bits = 16;
				break;
			
			case TOK_INT:
				cbt.bits = 32;
				break;
			
			case TOK_LONG:
				cbt.bits = 64;
				break;
			
			case TOK_USHORT:
				cbt.bits = 16;
				cbt.flags.field.Unsigned = 1;
				break;
			
			case TOK_UINT:
				cbt.bits = 32;
				cbt.flags.field.Unsigned = 1;
				break;
			
			case TOK_ULONG:
				cbt.bits = 64;
				cbt.flags.field.Unsigned = 1;
				break;
			
			case TOK_FSHORT:
				cbt.bits = 32;
				cbt.flags.field.Float = 1;
				break;
			
			case TOK_FLONG:
				cbt.bits = 64;
				cbt.flags.field.Float = 1;
				break;
			
			default:
				fprintf(stderr, "token %d\n", token);
				BUG();
		}
		
		type = BasicType::declare(sl->front(), NULL, cbt);
	}
	
	return type;
}

void declare_data(NameSpace *ns, StrList *ids, StrList *type,
                  Array *array, StrList *attr)
{
	bool is_inline = false;
	bool is_immutable = false;
	
	if (attr) {
		for (StrList::iterator i = attr->begin(); i != attr->end(); ++i) {
			const String *this_attr = *i;
	
			switch (this_attr->token) {
				case TOK_INLINE:
					is_inline = true;
					break;
				
				case TOK_IMMUTABLE:
					is_immutable = true;
					break;
				
				default:
					yyerrorfl(this_attr->file, this_attr->line,
					          "Invalid attribute \"%s\"", (*i)->c_str());
			}
		}
	}
	
	for (StrList::iterator i = ids->begin(); i != ids->end(); ++i) {
		try {
			Datum *d = Datum::declare(*i, ns, type, array);
			
			if (is_inline)
				d->set_inline();
			
			if (is_immutable)
				d->set_immutable();
		}
	
		catch (UserError) {
			// An error has already been displayed, but try to
			// continue and find more errors.
		}
	}
}

void declare_aliases(NameSpace *ns, StrList *ids, StrList *type,
                     bool is_typedef)
{
	StrList::iterator i;

	for (i = ids->begin(); i != ids->end(); ++i) {
		try {
			if (is_typedef)
				TypeDef::declare(*i, ns, type);
			else
				Alias::declare(*i, ns, type);
		}
	
		catch (UserError) {
			// An error has already been displayed, but try to
			// continue and find more errors.
		}
	}
}

void declare_basictypes(NameSpace *ns, StrList *ids, 
                        BasicType *type, bool is_typedef)
{
	assert(!type->def.flags.field.TypeDef);
	StrList::iterator i;
	
	for (i = ids->begin(); i != ids->end(); ++i) {
		try {
			BasicType *bt =
				BasicType::declare(*i, ns, type->def);
			
			bt->def.flags.field.TypeDef = is_typedef;
		}
	
		catch (UserError) {
			// An error has already been displayed, but try to
			// continue and find more errors.
		}
	}
}

Array::Array(NameSpace *LOOKUP_CTX)
{
	lookup_ctx = LOOKUP_CTX;
	
	for (int i = 0; i < 2; i++) {
		cons[i].con.ucon = 0;
		cons[i].type = TOK_UCON;
	}
}

void Array::set_unbounded()
{
	for (int i = 0; i < 2; i++)
		cons[i].type = TOK_NONE;
}

void Array::final_analysis()
{
	for (int i = 0; i < 2; i++) {
		if (cons[i].type == TOK_NONE) {
			if (i == 0)
				ca.bounds[0] = 0;
			else
				ca.bounds[1] = -1;
			
			continue;
		}
	
		if (cons[i].type == TOK_DCON) {
			Symbol *sym = lookup_sym(toplevel, dcons[i], lookup_ctx);
			datums[i] = dynamic_cast<Datum *>(sym);
			if (!datums[i]) {
				const String *str = dcons[i]->back();
				yyerrorfl(str->file, str->line, "\"%s\" is not a Datum.",
				          sym->get_fq_name()->flatten()->c_str());
				continue;
			}
			
			ca.bounds[i] = datums[i]->get_ucon(dcons[i]->back());
			cons[i].type = datums[i]->con_type;
		} else {
			ca.bounds[i] = cons[i].con.ucon;
		}
		
		if (cons[i].type == TOK_FCON) {
			yyerrorfl(strs[i]->file, strs[i]->line, 
			          "\"%s\" is not an integral Datum.",
			          strs[i]->c_str());
			throw UserError();
		}

		if (ca.bounds[i] < 0) {
			yyerrorfl(strs[i]->file, strs[i]->line, 
			          "\"%s\" %s.", strs[i]->c_str(),
			          cons[i].type == TOK_UCON ?
			          "does not fit in a signed 64-bit integer" :
			          "is negative");
			throw UserError();
		}
	}
	
	if (ca.bounds[1] >= 0 && ca.bounds[0] > ca.bounds[1]) {
		yyerrorfl(strs[0]->file, strs[0]->line,
		          "\"%s\" is larger than \"%s\".",
		          strs[0]->c_str(), strs[1]->c_str());
		throw UserError();
	}
}

void Array::set_bound(Con &con, int bound)
{
	assert(current_pass == 1);

	if (con.type == TOK_FCON) {
		yyerrorf("Array limits must be integers.");
		throw UserError();
	}
	
	if (con.type == TOK_ICON && con.con.icon <= 0) {
		yyerrorf("Array limits must be positive");
		throw UserError();
	}

	if (con.type == TOK_UCON && con.con.icon <= 0) {
		yyerrorf("Array limits must fit in a signed 64-bit integer");
		throw UserError();
	}
	
	cons[bound] = con;
	
	if (con.type == TOK_DCON) {
		assert(con.con.dcon);
		dcons[bound] = con.con.dcon;
		strs[bound] = con.con.dcon->flatten();
	} else {
		std::ostringstream ss(std::ostringstream::out);
		
		switch (con.type) {
			case TOK_FCON:
				ss << con.con.fcon;
				break;
			
			case TOK_UCON:
				ss << con.con.ucon;
				break;
			
			case TOK_ICON:
				ss << con.con.icon;
				break;
			
			case TOK_NONE:
				if (bound == 0)
					ss << "0";

				break;
			
			default:
				BUG();
		}
		
		strs[bound] = new String(ss.str().c_str(), cur_input_file, 
		                         curline, con.type);
	}
}


void UserNameSpace::output_lang(LangCallback *lcb, int arg1, void *arg2)
{
	lcb->output(this, arg1, arg2);
}

void BasicType::output_lang(LangCallback *lcb, int arg1, void *arg2)
{
	lcb->output(this, arg1, arg2);
}

void Interface::output_lang(LangCallback *lcb, int arg1, void *arg2)
{
	lcb->output(this, arg1, arg2);
}

void Struct::output_lang(LangCallback *lcb, int arg1, void *arg2)
{
	lcb->output(this, arg1, arg2);
}

void BitField::output_lang(LangCallback *lcb, int arg1, void *arg2)
{
	lcb->output(this, arg1, arg2);
}

void Enum::output_lang(LangCallback *lcb, int arg1, void *arg2)
{
	lcb->output(this, arg1, arg2);
}

void Alias::output_lang(LangCallback *lcb, int arg1, void *arg2)
{
	lcb->output(this, arg1, arg2);
}

void TypeDef::output_lang(LangCallback *lcb, int arg1, void *arg2)
{
	lcb->output(this, arg1, arg2);
}

void Datum::output_lang(LangCallback *lcb, int arg1, void *arg2)
{
	if (def.flags.field.Const)
		lcb->output(this, arg1, arg2);
}

void Method::output_lang(LangCallback *lcb, int arg1, void *arg2)
{
}

void Param::output_lang(LangCallback *lcb, int arg1, void *arg2)
{
}

int Interface::all_supers_traversal;
