#ifndef _UTIL_LIST_H
#define _UTIL_LIST_H

#include <assert.h>

namespace Util
{
	// Linked lists with nodes embedded in the data.  
	//
	// Duplicating for auto-init is ugly, but at least it's implementation
	// ugliness; inheriting would cause interface ugliness, with users
	// required to either declare ListAutoInit for the vastly common case,
	// or else use ListNoAutoInit pointers when following prev or next.
	//
	// Maybe some sort of templatization could be done.

	struct ListNoAutoInit {
		ListNoAutoInit *prev, *next;
		
		void init()
		{
			prev = next = this;
		}
		
		bool empty()
		{
			return next == this;
		}
		
		void del()
		{
			prev->next = next;
			next->prev = prev;
			init();
		}
		
		void add_front(ListNoAutoInit *newelem)
		{
			assert(newelem->empty());
		
			newelem->prev = this;
			next->prev = newelem;
			newelem->next = next;
			next = newelem;
		}
		
		void add_back(ListNoAutoInit *newelem)
		{
			assert(newelem->empty());

			newelem->next = this;
			prev->next = newelem;
			newelem->prev = prev;
			prev = newelem;
		}

		template<typename T, int offset>
		T *entry()
		{
			return reinterpret_cast<T *>(reinterpret_cast<ulong>(this) - offset);
		}
	};

	struct List {
		List *prev, *next;
		
		List()
		{
			init();
		}
		
		void init()
		{
			prev = next = this;
		}
		
		bool empty()
		{
			return next == this;
		}
		
		void del()
		{
			prev->next = next;
			next->prev = prev;
			init();
		}
		
		void add_front(List *newelem)
		{
			assert(newelem->empty());
		
			newelem->prev = this;
			next->prev = newelem;
			newelem->next = next;
			next = newelem;
		}
		
		void add_back(List *newelem)
		{
			assert(newelem->empty());

			newelem->next = this;
			prev->next = newelem;
			newelem->prev = prev;
			prev = newelem;
		}

		template<typename T, int offset>
		T *entry()
		{
			return reinterpret_cast<T *>(reinterpret_cast<ulong>(this) - offset);
		}
	};

	// Ick.  We can't pass in "member" as a template parameter, and
	// it'd be annoying to have to manually use offsetof all over
	// the place.  Use this as if it were a member function, but
	// be careful of the global name.
	
	#define listentry(T, member) entry<T, offsetof(T, member)>()
}

#endif
