#ifndef _KERN_PAGEALLOC_H
#define _KERN_PAGEALLOC_H

#include <System/Mem.h>

#include <kern/kernel.h>
#include <kern/list.h>
#include <kern/mem.h>

#include <arch/paging.h>
#include <arch/addrs.h>

#include <lowlevel/atomic.h>
#include <util/lock.h>

namespace Mem {
	class PageAllocZone;
	class Page;

	union PageFlags {
		struct bits {
			enum {
				Dirty = 0
			};
		};
	
		struct {
			uint32_t Dirty:1;
		};
		
		uint32_t pad;
	};

	// This is an array of all pages in all zones.  For now, zones must be
	// contiguous (or at least close enough to each other that it's not
	// a big deal to waste the space for the intervening page structs).
	
	extern Page *pages, *last_page;

	static inline uintptr_t page_to_phys(Page *page);
	
	static inline bool is_phys_page(Page *page)
	{
		return page >= Mem::pages && page < last_page;
	}
	
	struct Page {
		// A page is reserved if it is neither Free or InUse.
		// Free and InUse may not be set at the same time.
		
		static const u32 Free = 0x00000001;
		static const u32 InUse = 0x00000002;
		
		u32 flags;
		PageAllocZone *zone;

		union {
			// These fields are valid when a page is free, if it is the
			// first or last page of the chunk.
			
			struct {
				// For the first page of a chunk, this points to the last
				// page of the chunk.  For the last page of the chunk, this
				// points to the first page of the chunk.  For a single-page
				// chunk, this points to itself.  For a bin's list head, this
				// points to the smallest bin at least as large as itself
				// that has an available chunk.
				
				Page *otherside;
			} free;

			// These fields are valid when a page is in use (not free or
			// reserved).

			struct {
				int32_t refcount;
				PageFlags flags;
			} inuse;
		};

		// chunk when free, rmap when in use
		//
		// chunk points to the last page of the previous chunk in
		// the bin and the first page of the next chunk in the bin. 
		// Only valid for the first page of a chunk.  Prevchunk of the
		// first chunk and nextchunk of the last chunk point to the
		// bin's list head.
		//
		// rmap points to each mapping (as an RMapNode) of this page.
		// NOTE: rmap list currently unused
			
		Util::List chunk_rmap_list;
		
		void retain()
		{
			if (_UTIL_ASSERT_LEVEL >= 1) {
				if (!is_phys_page(this) || !(flags & InUse)) {
					in_fault++;
					printf("Page %p (phys %lx) retained flags %x, is_phys %d\n",
					       this, page_to_phys(this), flags,
					       (int)is_phys_page(this));
					__builtin_trap();
				}
			}
//			assert(flags & InUse);
			assert(inuse.refcount >= 1);
			ll_atomic_inc(&inuse.refcount);
		}

		void free_page();

		void release()
		{
			if (_UTIL_ASSERT_LEVEL >= 1) {
				if (!is_phys_page(this) || !(flags & InUse)) {
					in_fault++;
					printf("Page %p (phys %lx) released flags %x, is_phys %d\n",
					       this, page_to_phys(this), flags,
					       (int)is_phys_page(this));
					__builtin_trap();
				}
			}
//			assert(flags & InUse);
			assert(inuse.refcount >= 1);
			if (ll_atomic_dec_and_test(&inuse.refcount))
				free_page();
		}
		
		int32_t get_refcount()
		{
			assert(is_phys_page(this));
			assert(flags & InUse);
			assert(inuse.refcount >= 1);
			return inuse.refcount;
		}
	};

	static inline Page *phys_to_page(uintptr_t phys)
	{
		return pages + (phys >> Arch::page_shift);
	}
	
	static inline uintptr_t page_to_phys(Page *page)
	{
		return (page - pages) << Arch::page_shift;
	}
	
	static inline Page *kvirt_to_page(void *kvirt)
	{
		return phys_to_page(kvirt_to_phys(kvirt));
	}
	
	static inline void *page_to_kvirt(Page *page)
	{
		return phys_to_kvirt(page_to_phys(page));
	}

	class PageAllocZone {
		Lock::SpinLock lock;
	
		enum {
			num_bins = CONF_MEM_MAX_PAGE_ALLOC_BITS
		};
	
		// List head for each bin
		Page bins[num_bins];
		
		// First and last pages of the zone
		intptr_t start, end;
		size_t zonesize; // end - start + 1

		size_t chunk_size(Page *start);

		uint bin_to_size(int bin);
		int size_to_bin_alloc(size_t size);
		int size_to_bin_free(size_t size);

		Page *bin_to_head(int bin);

		void remove_chunk(Page *start, int bin);
		void add_to_bin(Page *chunk, int bin);
		Page *shrink_chunk(Page *start, int num_pages,
		                   size_t chunk_size, int bin);
		
	public:
		// base and size are in pages; all pages must be reserved
		// (i.e. flags set to zero).
		
		void init(uintptr_t base, size_t size);
	
		Page *alloc(uint num_pages);
		void free(Page *head, size_t num_pages);
	};
	
	class PageAlloc {
	public:
//		#include "mem/pagealloc-server/Mem/PageAlloc.h"
		
		// Architectures must define lists of zones, in preference order,
		// to use for each type of allocation; these are indices into
		// Arch::pagezonelists.
		//
		// The ISA DMA zonelist is for ISA DMA buffers, which must be below
		// some platform-dependent limit (typically 16MiB).
		//
		// The DMA32 zonelist is for devices on a 32-bit bus (such as
		// ordinary PCI), so that buffers over 4GiB are not used unless
		// the platform provides a suitable mapping mechanism (such
		// as an IOMMU).
		//
		// An architecture may define additional zonelists for its internal
		// use, but it's best to pass them directly as pointers than to
		// define additional numbers, so as to avoid conflict if the
		// generic list expands in the future.
		
		enum {
			zonelist_normal = 0,
			zonelist_dma32 = 1,
			zonelist_isadma = 2,
		};
		
		// Note that while num_pages does not have to be a power of 2 (and
		// the allocation size will not be rounded up to the next power of
		// 2), a request that is not a power of two may fail despite the
		// existence of a suitable chunk if there is no available chunk of
		// the next-higher power of 2.  num_pages may not be 0.
		
		static Page *alloc(uint num_pages, PageAllocZone *const *zonelist);

		static Page *alloc(uint num_pages, int zone = zonelist_normal)
		{
			assert(zone >= 0 && zone <= 2);
		
			return alloc(num_pages, Arch::pagezonelists[zone]);
		}
		
		// Any span of allocated pages may be freed; it does not have to
		// correspond to the size and start of the original allocation, and
		// it may be larger than the maximum allocation size (though this
		// is likely only useful during bootup when adding new chunks of
		// memory).  All pages must be in the same zone.  num_pages may not
		// be 0.
		
		static void free(Page *head, size_t num_pages)
		{
			head->zone->free(head, num_pages);
		}
	};
	
	extern PageAlloc page_alloc;

	
	static inline void *alloc_pages(int num,
	                                int zone = PageAlloc::zonelist_normal)
	{
		return Mem::page_to_kvirt(PageAlloc::alloc(num, zone));
	}

	static inline void free_pages(void *addr, int num)
	{
		if (addr)
			PageAlloc::free(kvirt_to_page(addr), num);
	}
	
	static inline void retain_if_phys(ulong addr)
	{
		Page *page = phys_to_page(addr);

		if (is_phys_page(page))
			page->retain();
	}

	static inline void release_if_phys(ulong addr)
	{
		Page *page = phys_to_page(addr);

		if (is_phys_page(page))
			page->release();
	}
}

#endif
